# Designed May. 2011 by Pierreden
# Permission to use, copy, modify this software for
# any non-commercial purpose is hereby granted

module LiveIvy
  class Branch
    attr_reader :nodes_array, :is_alive, :parents, :origin_node, :leafs_array

    def seed;
      IvyTool.seed;
    end

    def initialize(origin_node, parents)
      @nodes_array = []
      @leafs_array = []
      @is_alive    = true
      @parents     = parents
      if origin_node.class == LiveIvy::Node
        @origin_node                = origin_node
        @origin_node.primary_vector = Geom::linear_combination(seed.params[:branch][:spread], @origin_node.primary_vector.cross(@origin_node.adhesion_vector).transform([1,-1][rand(2).floor]), 1-seed.params[:branch][:spread], @origin_node.primary_vector)
      elsif origin_node.class == Geom::Point3d
        @origin_node = LiveIvy::Node.new(origin_node, self)
      end
      @nodes_array.push(@origin_node)
      IvyTool::branches.push(self)
    end

    def grow_stem
      if @nodes_array.last.float_length > seed.params[:branch][:floatlength]
        @is_alive = false
      end
      return false unless @is_alive
      @nodes_array.push(LiveIvy::Node.new(@nodes_array.last, self))
      total_length    = @nodes_array.last.length
      max_node_length = 0.63662*total_length*Math::acos(seed.params[:branch][:probability]+seed.params[:branch][:pattern]-1)
      return false if @parents == 3
      #puts " / #{max_n3} / #{max_n4}"
      for node in @nodes_array
        if node.length > max_node_length
          #seed.nodes_skipped += 1
          next
        end
        next if node.is_branched

        #seed.node_tests += 1
        weight = 1.0 - ((1*seed.params[:branch][:pattern])-Math::cos(node.length/total_length*2*3.14159265*0.5*0.5)).abs
        next if weight < seed.params[:branch][:probability]
        if rand*weight>seed.params[:branch][:probability]
          LiveIvy::Branch.new(node, @parents+1)
          node.is_branched = true
          break
        end
      end
    end

    def grow_leafs
      branch_length = @nodes_array.last.length
      #calculate_smooth_adhesion_vector

      1.times do
        for node in @nodes_array
          weight    = (node.length/branch_length)**0.7
          groundIvy = [0, -Geom::Vector3d.new(0, 0, 1).dot(node.adhesion_vector.normalize)].max
          weight    += groundIvy*(1-(node.length/branch_length)) #**2
          if rand*weight>seed.params[:leaf][:probability]
            leaf = LiveIvy::IvyLeaf.new(node, weight)
            @leafs_array.push(leaf)
          end
        end
      end
    end

    def point_sequens
      point_array = @nodes_array.select { |n| n.origin_node && n.origin_node.position_point }.map { |n| [n.origin_node.position_point, n.position_point] }.flatten
    end

    def calculate_smooth_adhesion_vector
      gaussian = [1, 2, 4, 7, 9, 10, 9, 7, 4, 2, 1]
      for g in 0..1
        for node_i in 0..@nodes_array.length-1
          next unless @nodes_array[node_i].smooth_adhesion_vector.nil?
          e = Geom::Vector3d.new(0, 0, 0)
          for i in -5..5
            tmpAdhesion = @nodes_array.first.adhesion_vector if node_i+i < 0
            tmpAdhesion = @nodes_array.last.adhesion_vector if node_i+i > @nodes_array.length-1
            tmpAdhesion = @nodes_array[node_i+i].adhesion_vector if node_i+i <= @nodes_array.length-1 && node_i+i >= 0
            e = e + Pdn::Vector3d.*(tmpAdhesion, gaussian[i+5])
          end
          e                                           = Geom::Vector3d.new(e.x/56.0, e.y/56.0, e.z/56.0)
          @nodes_array[node_i].smooth_adhesion_vector = e
        end
      end
    end
  end
end