# Designed May. 2011 by Pierreden
# Permission to use, copy, modify this software for
# any non-commercial purpose is hereby granted

module LiveIvy
  class Node
    @@nodes = []

    def self.nodes_count
      @@nodes.length
    end
    def self.reset_nodes
      @@nodes = []
    end

    attr_reader :origin_node, :position_point, :sum_vector,  :is_climbing, :climbing_normal, :length, :float_length, :legacy_vector, :branch
    attr_accessor :is_branched, :smooth_adhesion_vector, :primary_vector

    include Pdn::LiveVector::Adhesive
    include Pdn::LiveVector::Collision

    def initialize(origin_node, parent_branch)
      if origin_node.class == LiveIvy::Node
        @origin_node    = origin_node
        @adhesion_vector = [0, 0, 0]
        @branch         = parent_branch
        @is_climbing    = false
        @sum_vector     = sum_vector
        @position_point = @origin_node.position_point.offset(@sum_vector)
        @primary_vector = (Pdn::Vector3d.*(@origin_node.primary_vector, 0.5)+Pdn::Vector3d.*(@sum_vector-@external_vector, 0.5)).normalize
        @length         = @origin_node.length + @position_point.distance(@origin_node.position_point)
        @float_length   = @is_climbing ? 0.0 : @origin_node.float_length+@position_point.distance(@origin_node.position_point)
      else
        @origin_node     = nil
        @branch          = parent_branch
        @is_climbing     = false
        @is_branched     = false
        @position_point  = origin_node
        @length          = 0.0
        @float_length    = 0.0
        @primary_vector  = [0, 0, 1]
        @adhesion_vector = [0, 0, 0]
      end
      @@nodes.push(self)
    end

    def legacy_vector
      @origin_node.primary_vector
    end

    def random_vector
      @random_vector = Geom::Vector3d.new(rand-0.5, rand-0.5, rand-0.5+0.2).normalize
    end

    def gravity_vector
      @gravity_vector = Pdn::Vector3d.*(seed.params[:gravity][:vector], seed.params[:gravity][:weight]*seed.params[:branch][:nodesize])
      @gravity_vector = Pdn::Vector3d.*(@gravity_vector, (@origin_node.float_length/seed.params[:branch][:floatlength])**0.7)#*(1+((@branch.parents+1)/4)))
    end

    def wind_vector
      vector       = Sketchup.active_model.shadow_info["SunDirection"].reverse!.normalize
      @wind_vector = Pdn::Vector3d.*(vector, seed.params[:wind][:weight]*seed.params[:branch][:nodesize])
      @wind_vector = Pdn::Vector3d.*(@wind_vector, (@origin_node.float_length/seed.params[:branch][:floatlength])*(1+((@branch.parents+1)/4)))
    end

    def external_vector
      @external_vector = gravity_vector+wind_vector
    end

    def grow_vector
      @grow_vector = Pdn::Vector3d.*((Pdn::Vector3d.*(random_vector, seed.params[:random][:weight])+Pdn::Vector3d.*(legacy_vector, seed.params[:legacy][:weight])+Pdn::Vector3d.*(calculate_adhesion, seed.params[:adhesion][:weight])), seed.params[:branch][:nodesize])+cursor_vector
    end

    def sum_vector
      return @sum_vector if @sum_vector
      sum_vector  = grow_vector+external_vector
      @sum_vector = sum_vector + collision_adjustment_vector(sum_vector)
    end
    def cursor_vector
      cv = Geom::Vector3d.new(0,0,0)
      return cv unless seed.follow_cursor
      if seed.input_point.valid? && seed.input_point.position.class == Geom::Point3d
        #dist = origin_node.position_point.distance(seed.input_point.position)
        #ratio = 1-((dist.to_f-)/seed.follow_cursor[1].to_f)
        #ratio = 0 if ratio < 0
        cv = origin_node.position_point.vector_to(seed.input_point.position).normalize
      end
    end

    def force_inspect
      if @origin_node
        puts "Grow: Random:#{random_vector} / Legacy:#{legacy_vector} / Adhesion:#{adhesion_vector}"
        puts "External: Gravity:#{gravity_vector} / Wind:#{wind_vector}"
      end
    end

    def seed
      IvyTool.seed
    end


  end
end